#include "jabber/jlib.h"

thread tstream_new(int fd, tstream_onRead evt, void *evt_data)
{
    thread t;
    tstream ts;
    
    log(LOG_DEBUG,"tstream_new",NULL);

    t = thread_new();
    ts = pmalloc(t->p, sizeof(_tstream));
    ts->fd = fd;
    ts->queue = NULL;
    ts->read_evt = evt;
    ts->read_evt_data = evt_data;
    ts->data_aux1 = NULL;
    ts->data_aux2 = NULL;
    ts->data_aux3 = NULL;
    ts->data_aux4 = NULL;
    ts->data_aux5 = NULL;
    t->data = (void *)ts;
    t->id = pth_spawn(PTH_ATTR_DEFAULT, _tstream_main, (void *)t);
    pth_yield(NULL); /* let pth start the thread */
    
    return t;
}

int tstream_write(thread t, void* buffer, int bufsz)
{
    tstream ts = (tstream)(t->data);
    struct tstream_data *d_new, *d_count;

    if(buffer == NULL || bufsz == 0)
        return 0;

    log(LOG_DEBUG,"tstream_write",NULL);
    log(LOG_INSANE,"tstream_write_data",buffer);

/*    if(!pth_raise(t->id,0))
    {
        log(LOG_ERROR,"tstream_write","invalid thread");
        return -1;
    }*/
    d_new = malloc(sizeof(struct tstream_data));

    if(bufsz < 0) /* convience, send a -1 size and it will figure it out for you */
        bufsz = strlen(buffer);

    d_new->data = j_strdup(buffer);
    d_new->data_sz = bufsz;
    d_new->sent = 0;
    d_new->next = NULL;

    if(ts->queue == NULL)
    {
        ts->queue = d_new;
    }else{
        d_count = ts->queue;
        while(d_count != NULL)
        {
            if(d_count->next == NULL)
            {
                d_count->next = d_new;
                break;
            }
            d_count = d_count->next;
        }
    }

/*printf("XXX sending data to %d\n",t->id);*/
    pth_raise(t->id, SIGUSR2);
    if(t->id != pth_self())
        pth_yield(t->id); /* should allow the other thread to write to the socket right away */
}

static void *_tstream_main(void *arg)
{
    thread t = (thread)arg;
    tstream ts = (tstream)(t->data);
    pth_event_t sigevt, readevt, writeevt, evtring;
    sigset_t set;
    int sig, data_sz, write, wait;
    char data[1024];
    struct tstream_data *q;

    log(LOG_DEBUG,"_tstream_run","startup");
/*printf("XXX new tstream %d\n",t->id);*/
    
    pth_cleanup_push(_tstream_cleanup, (void *)t);
    
    sigemptyset(&set);
    sigaddset(&set, SIGUSR2);
    pth_sigmask(SIG_SETMASK, &set, NULL);
    sigevt = pth_event(PTH_EVENT_SIGS, &set, &sig);

    if(ts->read_evt == NULL)
        readevt = NULL; /* no even to handle reads, hence no reads */
    else
        readevt = pth_event(PTH_EVENT_FD|PTH_UNTIL_FD_READABLE, ts->fd);

    writeevt = pth_event(PTH_EVENT_FD|PTH_UNTIL_FD_WRITEABLE, ts->fd);

    evtring = pth_event_concat(sigevt, readevt, NULL);
    while(pth_wait(evtring) > 0)
    {
/*printf("XXX pth_event in %d\n",t->id);*/
        if(pth_event_occurred(readevt))
        { /* discussion: better to loop fetching all data waiting on socket
           * or return control to pth and let it come back?
           * If we return control, other waiting threads get a chance to breathe
           * if we read as much as possible, faster throughput for fast connections?
           */
            log(LOG_INSANE,"_tstream_run","read event");
            do
            {
                data_sz = pth_read(ts->fd, data, 1023);
                if(data_sz > 0)
                {
                    data[data_sz] = '\0';
                    log(LOG_INSANE,"_tstream_run",data);
                    (ts->read_evt)(t, data, data_sz);
                }else{
                    log(LOG_ERROR,"tstream","error reading from socket");
                    pth_exit(NULL);
                }
            } while(data_sz == 1023);
        }
        if(pth_event_occurred(sigevt) || pth_event_occurred(writeevt) || ts->queue != NULL)
        { /* if we were signaled from another thread, waiting on a socket write, or have something in the queue from this thread */
            log(LOG_INSANE,"_tstream_run","write event");

            wait = 0;
            while(ts->queue != NULL && wait == 0)
            {
                data_sz = pth_write(ts->fd, ts->queue->data + ts->queue->sent, ts->queue->data_sz - ts->queue->sent);
                if(data_sz < 0)
                {
                    log(LOG_ERROR,"tstream","error writing to socket");
                    pth_exit(NULL);
                }
                if((data_sz + ts->queue->sent) == ts->queue->data_sz)
                {
                    /* this packets empty, crush and move on */
                    q = ts->queue->next;
                    free(ts->queue->data);
                    free(ts->queue);
                    ts->queue = q;
                }else{
                    ts->queue->sent += data_sz;
                    wait = 1; /* not all of the data was sent, socket must be full, come back later */
                }
            }
                        

            if(ts->queue == NULL)
                evtring = pth_event_concat(sigevt, readevt, NULL);
            else
                evtring = pth_event_concat(sigevt, writeevt, readevt, NULL); 
        }
    }
}


static void _tstream_cleanup(void *arg)
{
    thread t = (thread)arg;
    tstream ts = (tstream)(t->data);
    struct tstream_data *q;
    int sz;

    log(LOG_DEBUG,"tstream_cleanup",NULL);

    /* XXX do more cleanup here? */
    while(ts->queue != NULL)
    {
        sz = pth_write(ts->fd, ts->queue->data + ts->queue->sent, ts->queue->data_sz - ts->queue->sent);
        if(sz < 0)
        {
            /* data lost, couldn't write to socket */
            break;
        }
        if((sz + ts->queue->sent) == ts->queue->data_sz)
        {
            /* this packets empty, crush and move on */
            q = ts->queue->next;
            free(ts->queue->data);
            free(ts->queue);
            ts->queue = q;
        }else{
            /* data lost, couldn't write to socket */
            break;
        }
    }
                        
    close(ts->fd);
    thread_free(t);
}

